@@ -3,6 +3,7 @@
 # specifically for storing facts (Predicate instances).
 # ------------------------------------------------------------------------------
 
+import sys
 import io
 import operator
 import collections
@@ -36,8 +37,9 @@ __all__ = [
 #------------------------------------------------------------------------------
 
 #------------------------------------------------------------------------------
-# Support function for printing ASP facts
-#------------------------------------------------------------------------------
+# Support function for printing ASP facts: Note: _trim_docstring() is taken from
+# PEP 257 (modified for Python 3): https://www.python.org/dev/peps/pep-0257/
+# ------------------------------------------------------------------------------
 
 def _format_asp_facts(iterator,output,width):
     tmp1=""
@@ -50,6 +52,71 @@ def _format_asp_facts(iterator,output,width):
             tmp1 = tmp1 + " " + fstr if tmp1 else fstr
     if tmp1: print(tmp1,file=output)
 
+def _trim_docstring(docstring):
+    if not docstring:
+        return ''
+    # Convert tabs to spaces (following the normal Python rules)
+    # and split into a list of lines:
+    lines = docstring.expandtabs().splitlines()
+    # Determine minimum indentation (first line doesn't count):
+    indent = sys.maxsize
+    for line in lines[1:]:
+        stripped = line.lstrip()
+        if stripped:
+            indent = min(indent, len(line) - len(stripped))
+    # Remove indentation (first line is special):
+    trimmed = [lines[0].strip()]
+    if indent < sys.maxsize:
+        for line in lines[1:]:
+            trimmed.append(line[indent:].rstrip())
+    # Strip off trailing and leading blank lines:
+    while trimmed and not trimmed[-1]:
+        trimmed.pop()
+    while trimmed and not trimmed[0]:
+        trimmed.pop(0)
+    # Return a single string:
+    return '\n'.join(trimmed)
+
+def _endstrip(string):
+    if not string: return
+    nl=string[-1]=='\n'
+    tmp=string.rstrip()
+    return tmp + '\n' if nl else tmp
+
+def _format_docstring(docstring,output):
+    if not docstring: return
+    tmp=_trim_docstring(docstring)
+    tmpstr = "".join(_endstrip("%     " + l) for l in tmp.splitlines(True))
+    if tmpstr:
+        print("% Description:",file=output)
+        print(tmpstr,file=output)
+
+def _maxwidth(lines):
+    return max([len(l) for l in lines])
+
+def _format_commented(fm,out):
+    pm = fm.predicate.meta
+    docstring = _trim_docstring(fm.predicate.__doc__) \
+        if fm.predicate.__doc__ else ""
+    indent = "    "
+    if pm.arity == 0:
+        lines = [ "Unary predicate signature:", indent + pm.name ]
+    else:
+        params=[str(fa.name) for fa in pm]
+        lines = [ "Predicate signature:",
+                  indent + "{}({})".format(pm.name, ",".join(params)) ]
+    if docstring:
+        lines.append("Description:")
+        for l in docstring.splitlines():lines.append(indent + l)
+    bar = "-" * _maxwidth(lines)
+    lines.insert(0,bar)
+    lines.append(bar)
+    for l in lines:
+        tmp = l.rstrip()
+        if tmp: print("% {}".format(tmp),file=out)
+        else: print("%",file=out)
+    return
+
 #------------------------------------------------------------------------------
 # A FactBase consisting of facts of different types
 #------------------------------------------------------------------------------
@@ -308,8 +375,7 @@ class FactBase(object):
             for fm in self._factmaps.values():
                 if first: first=False
                 else: print("",file=out)
-                pm=fm.predicate.meta
-                print("% FactBase predicate: {}/{}".format(pm.name,pm.arity),file=out)
+                _format_commented(fm,out)
                 _format_asp_facts(fm.factset,out,width)
 
         data = out.getvalue()
@@ -465,6 +531,10 @@ class FactBase(object):
         self.symmetric_difference_update(other)
         return self
 
+    def __getstate__(self):
+        self._check_init()
+        return self.__dict__
+
 
     #--------------------------------------------------------------------------
     # Set functions
@@ -857,9 +927,12 @@ class QueryImpl(Query):
     #--------------------------------------------------------------------------
     # Internal function to test whether a function has been called and add it
     #--------------------------------------------------------------------------
-    def _check_join_called_first(self, name):
-        if self._qspec.join is None and len(self._qspec.roots) > 1:
-            raise ValueError("'join' must be specified before '{}'".format(name))
+    def _check_join_called_first(self, name,endpoint=False):
+        if self._qspec.join is not None or len(self._qspec.roots) == 1: return
+        if endpoint:
+            raise ValueError(("A query over multiple predicates is incomplete without "
+                              "'join' clauses connecting these predicates"))
+        raise ValueError("A 'join' clause must be specified before '{}'".format(name))
 
     #--------------------------------------------------------------------------
     # Add a join expression
@@ -885,6 +958,17 @@ class QueryImpl(Query):
         nqspec = self._qspec.newp(where=where)
         return QueryImpl(self._factmaps, nqspec)
 
+    #--------------------------------------------------------------------------
+    # Add an orderered() flag
+    #--------------------------------------------------------------------------
+    def ordered(self, *expressions):
+        self._check_join_called_first("ordered")
+        if self._qspec.getp("order_by",None) is not None:
+            raise ValueError(("Invalid query 'ordered' declaration conflicts "
+                              "with previous 'order_by' declaration"))
+        nqspec = self._qspec.newp(ordered=True)
+        return QueryImpl(self._factmaps, nqspec)
+
     #--------------------------------------------------------------------------
     # Add an order_by expression
     #--------------------------------------------------------------------------
@@ -892,6 +976,10 @@ class QueryImpl(Query):
         self._check_join_called_first("order_by")
         if not expressions:
             nqspec = self._qspec.newp(order_by=None)   # raise exception
+        elif self._qspec.getp("ordered",False):
+            raise ValueError(("Invalid query 'order_by' declaration '{}' "
+                              "conflicts with previous 'ordered' "
+                              "declaration").format(expressions))
         else:
             nqspec = self._qspec.newp(
                 order_by=process_orderby(expressions,self._qspec.roots))
@@ -975,7 +1063,7 @@ class QueryImpl(Query):
     # Select to display all the output of the query
     # --------------------------------------------------------------------------
     def all(self):
-        self._check_join_called_first("all")
+        self._check_join_called_first("all",endpoint=True)
 
         qe = QueryExecutor(self._factmaps, self._qspec)
         return qe.all()
@@ -984,7 +1072,7 @@ class QueryImpl(Query):
     # Show the single element and throw an exception if there is more than one
     # --------------------------------------------------------------------------
     def singleton(self):
-        self._check_join_called_first("singleton")
+        self._check_join_called_first("singleton",endpoint=True)
 
         qe = QueryExecutor(self._factmaps, self._qspec)
         found = None
@@ -997,7 +1085,7 @@ class QueryImpl(Query):
     # Return the count of elements - Note: the behaviour of what is counted
     # changes if group_by() has been specified.
     # --------------------------------------------------------------------------
-    def count(self):
+    def count(self,endpoint=True):
         self._check_join_called_first("count")
 
         qe = QueryExecutor(self._factmaps, self._qspec)
@@ -1017,7 +1105,7 @@ class QueryImpl(Query):
     #--------------------------------------------------------------------------
     # Return the first element of the query
     # --------------------------------------------------------------------------
-    def first(self):
+    def first(self,endpoint=True):
         self._check_join_called_first("first")
 
         qe = QueryExecutor(self._factmaps, self._qspec)
@@ -1026,7 +1114,7 @@ class QueryImpl(Query):
     #--------------------------------------------------------------------------
     # Delete a selection of fact
     #--------------------------------------------------------------------------
-    def delete(self):
+    def delete(self,endpoint=True):
         self._check_join_called_first("delete")
 
         qe = QueryExecutor(self._factmaps, self._qspec)
